from abc import abstractmethod
import logging
import re
from argparse import Namespace
from functools import lru_cache

from datasets import Dataset
from evaluate import load as load_metric
from llama_cpp import LlamaGrammar

from evoprompt.task.task import SYSTEM_MESSAGE, DatasetDatum, Task

logger = logging.getLogger(__name__)


def grammar_continuous_with_arbitrary_end(sequence: list[str], quote: str = '"'):
    if len(sequence) == 0:
        return ""

    if quote is None:
        quote = ""

    escape_item = lambda text: text.replace('"', '\\"')

    grammar = quote + escape_item(sequence[0]) + quote
    for idx, item in enumerate(sequence[1:]):
        new_item = "(" + quote + escape_item(item) + quote + ")?"
        if idx == 0:
            grammar += new_item
        else:
            grammar = grammar[: -2 * idx] + new_item + grammar[-2 * idx :]
    return grammar


# a grammar in GBNF notation which takes the context into account
# NOTE: for some reason the program crashes if the grammar is an instance attribute of the task class
@lru_cache
def extractive_qa_grammar_fn(context: str, verbose: bool = False):
    grammar_str = "(\n"
    tokens = re.split(r"(\W+)", context.strip())[:-1]

    for i in range(len(tokens)):
        if i > 0:
            grammar_str += "\n|\n"
        grammar_str += "(" + grammar_continuous_with_arbitrary_end(tokens[i:]) + ")"
    grammar_str += "\n)"
    return LlamaGrammar.from_string(f"root ::= {grammar_str}", verbose=verbose)


class QuestionAnswering(Task):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.metric = load_metric("squad")

    def predict(self, prompt: str, datum: DatasetDatum):
        # run model for inference
        grammar = None
        if self.use_grammar:
            # context-sensitive grammar
            context = self._get_context_from_datum(datum)
            try:
                grammar = extractive_qa_grammar_fn(context)
            except Exception as e:
                logger.exception(
                    "Could not create grammar for context (potentially maximum recursion depth exceeded), therefore we do not use a grammar for this sample.\nContext (with length %d): %s",
                    len(context),
                    context,
                    exc_info=e,
                )

        response, usage = self.model(
            system_message=SYSTEM_MESSAGE,
            prompt=prompt,
            prompt_appendix=self._get_prompt_text_for_datum(datum),
            grammar=grammar,
        )

        if not self.use_grammar:
            # we postprocess the model output to return as answer
            response = response.strip()

        return response, usage

    def _get_prompt_text_for_datum(self, datum: DatasetDatum) -> str:
        context = self._get_context_from_datum(datum)
        question = self._get_question_from_datum(datum)
        return (
            "\nContext: " + '"' + context + '"' + "\nQuestion: " + '"' + question + '"'
        )

    @abstractmethod
    def _get_id_from_datum(self, datum: DatasetDatum):
        pass

    @abstractmethod
    def _get_context_from_datum(self, datum: DatasetDatum):
        pass

    @abstractmethod
    def _get_question_from_datum(self, datum: DatasetDatum):
        pass

    @abstractmethod
    def _get_gold_labels_from_datum(self, datum: DatasetDatum):
        pass

    def _evaluate_sample(self, prompt: str, datum: DatasetDatum):
        _id = self._get_id_from_datum(datum)
        context = self._get_context_from_datum(datum)
        question = self._get_question_from_datum(datum)
        gold_answers = self._get_gold_labels_from_datum(datum)
        answer, usage = self.predict(prompt, datum)
        # input(f'*** PREDICTION ***\n\tContext: {datum["context"]}\n\tQuestion: {datum["question"]}\n\tAnswer: {answer}\n\tGold answer: {datum["answers"]["text"][0]}')
        # TODO check if answer is lower-cased in metric computation

        result = self.metric.compute(
            predictions=[{"prediction_text": answer, "id": _id}],
            references=[{"answers": gold_answers, "id": _id}],
        )
        return result["f1"] / 100, usage

    def _aggregate_result(self, results: list[float]) -> float:
        return sum(results) / len(results)

    def evaluate(
        self,
        prompt: str,
        dataset: Dataset,
        parent_histories: list[list[float]] | None = None,
    ):
        def replace_symbol_for_grammar(sample: DatasetDatum):
            symbol_replacement_mapping = {
                "\u2013": "-",
                "\u2014": "-",
            }
            symbol_replacement_mapping = dict(
                (re.escape(k), v) for k, v in symbol_replacement_mapping.items()
            )
            symbol_replacement_pattern = re.compile(
                "|".join(symbol_replacement_mapping.keys())
            )
            replace_fn = lambda text: symbol_replacement_pattern.sub(
                lambda m: symbol_replacement_mapping[re.escape(m.group(0))], text
            )
            sample["context"] = replace_fn(sample["context"])
            sample["answers"]["text"] = [
                replace_fn(text) for text in sample["answers"]["text"]
            ]
            return sample

        if self.use_grammar:
            # NOTE: the LlamaGrammar has issues with symbol '–' therefore we replace all occurences with '-' (hyphen)
            dataset = dataset.map(replace_symbol_for_grammar, desc="Replacing symbols")
        return super().evaluate(prompt, dataset, parent_histories=parent_histories)

    @property
    def metric_name(self):
        return "f1"

    @property
    def base_prompt(self):
        # TODO find good prompt
        return """In this task, you are given a context and a question. The task is to answer the question given the context. Return only the answer without any other text. Make sure that the answer is taken directly from the context."""


class SQuAD(QuestionAnswering):
    shorthand = "squad"

    def load_validation_set(
        self, validation_dataset: str | None, validation_split: str | None
    ):
        return super().load_validation_set("squad", "train[:200]")

    def load_test_set(self, test_dataset: str, test_split: str | None):
        return super().load_test_set("squad", "validation")

    def _get_context_from_datum(self, datum: DatasetDatum):
        return datum["context"]

    def _get_question_from_datum(self, datum: DatasetDatum):
        return datum["question"]

    def _get_id_from_datum(self, datum: DatasetDatum):
        return datum["id"]

    def _get_gold_labels_from_datum(self, datum: DatasetDatum):
        return datum["answers"]
