from abc import abstractmethod
import logging
from argparse import Namespace
from functools import lru_cache

from llama_cpp import LlamaGrammar

from evoprompt.task.task import SYSTEM_MESSAGE, DatasetDatum, Task

logger = logging.getLogger(__name__)


# a simple grammar in GBNF notation which allows either positive or negative
# NOTE: for some reason the program crashes if the grammar is an instance attribute of the task class
SA_OUTPUTS = set(("positive", "negative"))


@lru_cache
def sa_grammar_fn(verbose: bool = False):
    return LlamaGrammar.from_string(
        "root ::= ({})".format("|".join(('"' + cons + '"') for cons in SA_OUTPUTS)),
        verbose=verbose,
    )


class SentimentAnalysis(Task):
    def predict(self, prompt: str, datum: DatasetDatum):
        # run model for inference using grammar to constrain output
        # TODO grammar also depends on prompt and vice-versa -> what are good labels?
        response, usage = self.model(
            system_message=SYSTEM_MESSAGE,
            prompt=prompt,
            prompt_appendix=self._get_prompt_text_for_datum(datum),
            grammar=sa_grammar_fn() if self.use_grammar else None,
        )

        if not self.use_grammar:
            # we postprocess the model output to return as answer
            response = response.strip()
        # input(f"*** PREDICTION***\n\tText: {text}\n\tSentiment: {response}")

        return response, usage

    def _evaluate_sample(self, prompt: str, datum: DatasetDatum):
        gold_label = self._get_gold_label_for_datum(datum)
        sst2_labels = {"negative": 0, "positive": 1}
        response, usage = self.predict(prompt=prompt, datum=datum)
        response = response.lower()
        if self.use_grammar:
            # model output is from label space
            answer_label = sst2_labels[response]
        else:
            answer_label = None
            for label in sst2_labels.keys():
                if label in response:
                    answer_label = sst2_labels[label]
                    break
            else:
                logger.warning(f"Invalid answer: {response}")
                return "failed", usage

        classification_result = "incorrect" if answer_label != gold_label else "correct"
        return classification_result, usage
    
    def _get_prompt_text_for_datum(self, datum: DatasetDatum) -> str:
        return "\nInput: " + '"' + self._get_text_for_datum(datum) + '"'
    
    @abstractmethod
    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        pass

    @abstractmethod
    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        pass

    def _aggregate_result(self, results: list[str]) -> float:
        num_correct_results = sum(1 for result in results if result == "correct")
        accuracy = num_correct_results / len(results)
        return accuracy

    @property
    def metric_name(self):
        return "accuracy"

    @property
    def base_prompt(self):
        #  from the paper: RLPROMPT: Optimizing Discrete Text Prompts with Reinforcement Learning
        return """In this task, you are given sentences from movie reviews. The task is to classify a sentence as 'positive' if the sentiment of the sentence is positive or as 'negative' if the sentiment of the sentence is negative. Return label only without any other text."""


class SST2(SentimentAnalysis):
    shorthand = "sst2"

    def load_validation_set(
        self, validation_dataset: str | None, validation_split: str | None
    ):
        return super().load_validation_set("stanfordnlp/sst2", "validation[:200]")

    def load_test_set(self, test_dataset: str, test_split: str | None):
        return super().load_test_set("stanfordnlp/sst2", "test")

    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        return datum["sentence"]

    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        return datum["label"]
