import logging
import re
from abc import ABCMeta, abstractmethod
from argparse import Namespace
from functools import lru_cache
from statistics import mean
from typing import Union

from datasets import Dataset, load_dataset
from llama_cpp import LlamaGrammar, deque
from tqdm import tqdm

from evoprompt.models import LLMModel
from evoprompt.opt_types import ModelUsage
from evoprompt.utils import log_calls

logger = logging.getLogger(__name__)


SYSTEM_MESSAGE = """
You are given an instruction that describes a task, paired with an input that provides further context. Write a response that appropriately completes the request.
"""

DatasetDatum = dict


class EarlyStoppingMonitor:

    @abstractmethod
    def update(self, score: float) -> bool:
        raise NotImplementedError


class MomentBasedStopping(EarlyStoppingMonitor):
    """
    Watch the first derivative (moment) of the metric to determine when to stop.
    """

    def __init__(
        self,
        *,
        patience: int = 10,
        start_after: int = 20,
        min_moment_magnitude: float = 0.001,
    ):
        self.patience = patience
        self.start_after = start_after
        self.min_moment_magnitude = min_moment_magnitude

        self.moment_magnitudes = deque(maxlen=patience)
        self.last_score = 0.0
        self.num_calls = 0

    def update(self, score: float) -> bool:
        # caclulate the current moment (dx/dt)
        self.num_calls += 1
        if self.num_calls < self.start_after:
            return False

        self.moment_magnitudes.append(abs(score - self.last_score))
        self.last_score = score
        if len(self.moment_magnitudes) < self.patience:
            return False

        if mean(self.moment_magnitudes) < self.min_moment_magnitude:
            return True

        return False


class ParentBaselineBasedStopping(EarlyStoppingMonitor):

    def __init__(
        self,
        parent_histories: list[list[float]],
        *,
        patience: int = 10,
        start_after: int = 20,
        min_improvement: float = 0.001,
    ):
        self.parent_histories = parent_histories
        self.patience = patience
        self.start_after = start_after
        self.min_improvement = min_improvement
        self.num_calls = 0
        self.improvement_memory = deque(maxlen=patience)

    def update(self, score: float) -> bool:
        self.num_calls += 1
        if self.num_calls < self.start_after:
            return False

        parent_values = [  # get the metric value of the parents at the current step
            (
                parent_history[self.num_calls - 1]
                if len(parent_history) >= self.num_calls
                else parent_history[-1]  # extend with last value
            )
            for parent_history in self.parent_histories
        ]
        self.improvement_memory.append(
            score - max(parent_values)  # compare with the best parent
        )

        if len(self.improvement_memory) < self.patience:
            return False

        if max(self.improvement_memory) < self.min_improvement:
            # if the highest improvement is less than the minimum improvement, we stop
            return True

        return False


class Task(metaclass=ABCMeta):
    shorthand: str
    validation_dataset: Dataset
    test_dataset: Dataset

    def __init__(
        self,
        model: Union[LLMModel],
        validation_dataset: str | None = None,
        test_dataset: str | None = None,
        *,
        use_grammar: bool,
        no_early_stopping: bool = False,
        evaluate_shortest_first: bool = False,
        validation_split: str | None = None,
        test_split: str | None = None,
        debug: bool = False,
        **kwargs,
    ) -> None:
        self.model = model
        self.debug = debug
        # whether we use the grammar to constrain the model output or not
        self.use_grammar = use_grammar
        self.no_early_stopping = no_early_stopping
        self.evaluate_shortest_first = evaluate_shortest_first
        if evaluate_shortest_first and no_early_stopping:
            logger.warning(
                "Both 'evaluate_shortest_first' and 'no_early_stopping' are set. This makes no sense"
            )
        self.validation_dataset = self.load_validation_set(
            validation_dataset, validation_split
        )
        self.test_dataset = self.load_test_set(test_dataset, test_split)

    def load_validation_set(
        self, validation_dataset: str, validation_split: str | None
    ):
        dataset = load_dataset(validation_dataset, split=validation_split)
        if self.debug and len(dataset) > 5:
            dataset = dataset.select(range(5))
        return dataset

    def load_test_set(self, test_dataset: str, test_split: str | None):
        dataset = load_dataset(test_dataset, split=test_split)
        if self.debug and len(dataset) > 20:
            dataset = dataset.select(range(20))
        return dataset

    @abstractmethod
    def predict(self, prompt: str, *args, **kwargs) -> tuple[str, ModelUsage]:
        pass

    @abstractmethod
    def _evaluate_sample(
        self, prompt: str, datum: DatasetDatum
    ) -> tuple[str, ModelUsage]:
        pass

    @abstractmethod
    def _get_prompt_text_for_datum(self, datum: DatasetDatum) -> str:
        pass

    @abstractmethod
    def _aggregate_result(self, results: list) -> float:
        pass

    def evaluate(
        self,
        prompt: str,
        dataset: Dataset,
        parent_histories: list[list[float]] | None = None,
    ) -> tuple[float, ModelUsage, list[float]]:

        early_stopping: EarlyStoppingMonitor
        early_stopping_params = {
            "patience": max(len(dataset) // 20, 5),
            "start_after": max(len(dataset) // 5, 5),
        }
        if parent_histories is not None:
            early_stopping = ParentBaselineBasedStopping(
                parent_histories, **early_stopping_params
            )
        else:
            early_stopping = MomentBasedStopping(**early_stopping_params)

        results: list = []
        if self.evaluate_shortest_first:
            dataset = sorted(
                dataset, key=lambda x: len(self._get_prompt_text_for_datum(x))
            )
        dataset_iterator: tqdm[DatasetDatum] = tqdm(
            dataset, desc="evaluating prompt", leave=False
        )
        evaluation_usage = ModelUsage()
        evaluation_history = []

        for datum in dataset_iterator:
            result, usage = self._evaluate_sample(prompt, datum)
            results.append(result)
            current_metric = self._aggregate_result(results)
            dataset_iterator.set_postfix(
                {self.metric_name: f"{current_metric*100:.1f}%"}
            )
            evaluation_usage += usage
            evaluation_history.append(current_metric)
            if not self.no_early_stopping and early_stopping.update(current_metric):
                logger.info(
                    f"Early stopping after {len(results)} samples with {self.metric_name} of {current_metric*100:.1f}%"
                )
                break
        # input(f'F1 score: {current_metric:.2f}')

        return self._aggregate_result(results), evaluation_usage, evaluation_history

    @log_calls("Evaluating validation dataset")
    def evaluate_validation(
        self, prompt: str, parent_histories: list[list[float]] | None = None
    ):
        return self.evaluate(prompt, self.validation_dataset, parent_histories)

    @log_calls("Evaluating test dataset")
    def evaluate_test(self, prompt: str):
        return self.evaluate(prompt, self.test_dataset)

    @property
    @abstractmethod
    def metric_name(self) -> str:
        pass

    @property
    @abstractmethod
    def base_prompt(self) -> str:
        pass
